import pytest

from cachelib import BaseCache


@pytest.fixture(autouse=True)
def cache_factory(request):
    def _factory(self, *args, **kwargs):
        return BaseCache(*args, **kwargs)

    request.cls.cache_factory = _factory


class TestBaseCache:
    def test_get(self):
        cache = self.cache_factory()
        assert cache.get("bacon") is None

    def test_delete(self):
        cache = self.cache_factory()
        assert cache.delete("eggs")

    def test_get_many(self):
        cache = self.cache_factory()
        keys = ["bacon", "spam", "eggs"]
        expected = [None] * 3
        assert cache.get_many(*keys) == expected

    def test_get_dict(self):
        cache = self.cache_factory()
        keys = ["bacon", "spam", "eggs"]
        expected = dict.fromkeys(keys, None)
        assert cache.get_dict(*keys) == expected

    def test_set(self):
        cache = self.cache_factory()
        assert cache.set("sausage", "tomato")

    def test_add(self):
        cache = self.cache_factory()
        assert cache.add("baked", "beans")

    def test_set_many(self):
        cache = self.cache_factory()
        keys = ["bacon", "spam", "eggs"]
        mapping = dict.fromkeys(keys, None)
        assert cache.set_many(mapping) == keys

    def test_delete_many(self):
        cache = self.cache_factory()
        keys = ["bacon", "spam", "eggs"]
        assert cache.delete_many(*keys) == keys

    def test_has(self):
        cache = self.cache_factory()
        with pytest.raises(NotImplementedError):
            cache.has("lobster")

    def test_clear(self):
        cache = self.cache_factory()
        assert cache.clear()

    def test_inc(self):
        cache = self.cache_factory()
        assert cache.inc("crevettes", delta=10) == 10

    def test_dec(self):
        cache = self.cache_factory()
        assert cache.dec("truffle", delta=10) == -10



from time import sleep

import pytest
from conftest import TestData
from conftest import under_uwsgi


class CommonTests(TestData):
    """A base set of tests to be run for all cache types"""

    def test_set_get(self):
        cache = self.cache_factory()
        for k, v in self.sample_pairs.items():
            assert cache.set(k, v)
            assert cache.get(k) == v

    def test_set_get_many(self):
        cache = self.cache_factory()
        result = cache.set_many(self.sample_pairs)
        assert result == list(self.sample_pairs.keys())
        values = cache.get_many(*self.sample_pairs)
        assert values == list(self.sample_pairs.values())

    def test_get_dict(self):
        cache = self.cache_factory()
        cache.set_many(self.sample_pairs)
        d = cache.get_dict(*self.sample_pairs)
        assert d == self.sample_pairs

    def test_delete(self):
        cache = self.cache_factory()
        for k, v in self.sample_pairs.items():
            cache.set(k, v)
            assert cache.delete(k)
            assert not cache.get(k)

    def test_delete_many(self):
        cache = self.cache_factory()
        cache.set_many(self.sample_pairs)
        result = cache.delete_many(*self.sample_pairs)
        assert result == list(self.sample_pairs.keys())
        assert not any(cache.get_many(*self.sample_pairs))

    def test_add(self):
        cache = self.cache_factory()
        cache.set_many(self.sample_pairs)
        for k in self.sample_pairs:
            assert not cache.add(k, "updated")
        assert cache.get_many(*self.sample_pairs) == list(self.sample_pairs.values())
        for k, v in self.sample_pairs.items():
            assert cache.add(f"{k}-new", v)
            assert cache.get(f"{k}-new") == v

    def test_inc_dec(self):
        cache = self.cache_factory()
        for n in self.sample_numbers:
            assert not cache.get(f"{n}-key-inc")
            assert cache.inc(f"{n}-key-inc", n) == n
            assert cache.get(f"{n}-key-inc") == n
            assert cache.dec(f"{n}-key-dec", n) == -n
            assert cache.get(f"{n}-key-dec") == -n
            assert cache.dec(f"{n}-key-inc", 5) == n - 5

    def test_expiration(self):
        if under_uwsgi():
            pytest.skip(
                "uwsgi uses a separate sweeper thread to clean"
                " expired chache entries, thus the testing"
                " of such feature must be handled differently"
                " from other cache types."
            )
        cache = self.cache_factory()
        for k, v in self.sample_pairs.items():
            cache.set(f"{k}-t0", v, timeout=0)
            cache.set(f"{k}-t1", v, timeout=1)
        sleep(4)
        for k, v in self.sample_pairs.items():
            assert cache.get(f"{k}-t0") == v
            assert not cache.get(f"{k}-t1")



import pytest
from clear import ClearTests
from common import CommonTests
from has import HasTests

from cachelib import MemcachedCache


@pytest.fixture(autouse=True)
def cache_factory(request):
    def _factory(self, *args, **kwargs):
        mc = MemcachedCache(*args, **kwargs)
        mc._client.flush_all()
        return mc

    request.cls.cache_factory = _factory


@pytest.mark.usefixtures("memcached_server")
class TestMemcachedCache(CommonTests, ClearTests, HasTests):
    pass



import pytest
from clear import ClearTests
from common import CommonTests
from has import HasTests

from cachelib import RedisCache


@pytest.fixture(autouse=True)
def cache_factory(request):
    def _factory(self, *args, **kwargs):
        rc = RedisCache(*args, port=6360, **kwargs)
        rc._client.flushdb()
        return rc

    request.cls.cache_factory = _factory


@pytest.mark.usefixtures("redis_server")
class TestRedisCache(CommonTests, ClearTests, HasTests):
    pass



import pickle
import typing as _t

from cachelib.base import BaseCache


class RedisCache(BaseCache):
    """Uses the Redis key-value store as a cache backend.

    The first argument can be either a string denoting address of the Redis
    server or an object resembling an instance of a redis.Redis class.

    Note: Python Redis API already takes care of encoding unicode strings on
    the fly.

    :param host: address of the Redis server or an object which API is
                 compatible with the official Python Redis client (redis-py).
    :param port: port number on which Redis server listens for connections.
    :param password: password authentication for the Redis server.
    :param db: db (zero-based numeric index) on Redis Server to connect.
    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`~BaseCache.set`. A timeout of
                            0 indicates that the cache never expires.
    :param key_prefix: A prefix that should be added to all keys.

    Any additional keyword arguments will be passed to ``redis.Redis``.
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        password: _t.Optional[str] = None,
        db: int = 0,
        default_timeout: int = 300,
        key_prefix: _t.Optional[str] = None,
        **kwargs: _t.Any
    ):
        BaseCache.__init__(self, default_timeout)
        if host is None:
            raise ValueError("RedisCache host parameter may not be None")
        if isinstance(host, str):
            try:
                import redis
            except ImportError as err:
                raise RuntimeError("no redis module found") from err
            if kwargs.get("decode_responses", None):
                raise ValueError("decode_responses is not supported by RedisCache.")
            self._client = redis.Redis(
                host=host, port=port, password=password, db=db, **kwargs
            )
        self.key_prefix = key_prefix or ""

    def _normalize_timeout(self, timeout: _t.Optional[int]) -> int:
        timeout = BaseCache._normalize_timeout(self, timeout)
        if timeout == 0:
            timeout = -1
        return timeout

    def dump_object(self, value: _t.Any) -> bytes:
        """Dumps an object into a string for redis.  By default it serializes
        integers as regular string and pickle dumps everything else.
        """
        if isinstance(type(value), int):
            return str(value).encode("ascii")
        return b"!" + pickle.dumps(value)

    def load_object(self, value: _t.Optional[bytes]) -> _t.Any:
        """The reversal of :meth:`dump_object`.  This might be called with
        None.
        """
        if value is None:
            return None
        if value.startswith(b"!"):
            try:
                return pickle.loads(value[1:])
            except pickle.PickleError:
                return None
        try:
            return int(value)
        except ValueError:
            # before 0.8 we did not have serialization.  Still support that.
            return value

    def get(self, key: str) -> _t.Any:
        return self.load_object(self._client.get(self.key_prefix + key))

    def get_many(self, *keys: str) -> _t.List[_t.Any]:
        if self.key_prefix:
            prefixed_keys = [self.key_prefix + key for key in keys]
        else:
            prefixed_keys = [k for k in keys]
        return [self.load_object(x) for x in self._client.mget(prefixed_keys)]

    def set(
        self, key: str, value: _t.Any, timeout: _t.Optional[int] = None
    ) -> _t.Optional[bool]:
        timeout = self._normalize_timeout(timeout)
        dump = self.dump_object(value)
        if timeout == -1:
            result = self._client.set(name=self.key_prefix + key, value=dump)
        else:
            result = self._client.setex(
                name=self.key_prefix + key, value=dump, time=timeout
            )
        return result

    def add(self, key: str, value: _t.Any, timeout: _t.Optional[int] = None) -> bool:
        timeout = self._normalize_timeout(timeout)
        dump = self.dump_object(value)
        return self._client.setnx(
            name=self.key_prefix + key, value=dump
        ) and self._client.expire(name=self.key_prefix + key, time=timeout)

    def set_many(
        self, mapping: _t.Dict[str, _t.Any], timeout: _t.Optional[int] = None
    ) -> _t.List[_t.Any]:
        timeout = self._normalize_timeout(timeout)
        # Use transaction=False to batch without calling redis MULTI
        # which is not supported by twemproxy
        pipe = self._client.pipeline(transaction=False)

        for key, value in mapping.items():
            dump = self.dump_object(value)
            if timeout == -1:
                pipe.set(name=self.key_prefix + key, value=dump)
            else:
                pipe.setex(name=self.key_prefix + key, value=dump, time=timeout)
        results = pipe.execute()
        return [k for k, was_set in zip(mapping.keys(), results) if was_set]

    def delete(self, key: str) -> bool:
        return bool(self._client.delete(self.key_prefix + key))

    def delete_many(self, *keys: str) -> _t.List[_t.Any]:
        if not keys:
            return []
        if self.key_prefix:
            prefixed_keys = [self.key_prefix + key for key in keys]
        else:
            prefixed_keys = [k for k in keys]
        self._client.delete(*prefixed_keys)
        return [k for k in prefixed_keys if not self.has(k)]

    def has(self, key: str) -> bool:
        return bool(self._client.exists(self.key_prefix + key))

    def clear(self) -> bool:
        status = 0
        if self.key_prefix:
            keys = self._client.keys(self.key_prefix + "*")
            if keys:
                status = self._client.delete(*keys)
        else:
            status = self._client.flushdb()
        return bool(status)

    def inc(self, key: str, delta: int = 1) -> _t.Optional[int]:
        return self._client.incr(name=self.key_prefix + key, amount=delta)

    def dec(self, key: str, delta: int = 1) -> _t.Optional[int]:
        return self._client.incr(name=self.key_prefix + key, amount=-delta)



# type: ignore
import inspect

import pytest

from cachelib import BaseCache
from cachelib import FileSystemCache
from cachelib import MemcachedCache
from cachelib import RedisCache
from cachelib import SimpleCache


@pytest.fixture(autouse=True)
def create_cache_list(request, tmpdir):
    mc = MemcachedCache()
    mc._client.flush_all()
    rc = RedisCache(port=6360)
    rc._client.flushdb()
    request.cls.cache_list = [FileSystemCache(tmpdir), mc, rc, SimpleCache()]


@pytest.mark.usefixtures("redis_server", "memcached_server")
class TestInterfaceUniformity:
    def test_types_have_all_base_methods(self):
        public_api_methods = [
            meth
            for meth in inspect.getmembers(BaseCache, predicate=inspect.isfunction)
            if not meth[0].startswith("_")
        ]
        for cache_type in self.cache_list:
            for meth in public_api_methods:
                assert hasattr(cache_type, meth[0]) and callable(meth[1])



import pytest
from clear import ClearTests
from common import CommonTests
from has import HasTests

from cachelib import UWSGICache


@pytest.fixture(autouse=True)
def cache_factory(request):
    def _factory(self, *args, **kwargs):
        uwc = UWSGICache(*args, **kwargs)
        uwc.clear()
        return uwc

    request.cls.cache_factory = _factory


class TestUwsgiCache(CommonTests, ClearTests, HasTests):
    pytest.importorskip(
        "uwsgi",
        reason="could not import 'uwsgi'. Make sure to "
        "run pytest under uwsgi for testing UWSGICache",
    )
